% This script shows how to solve the neoclassical growh model by perturbation.
%
% The model is defined by:
%
% Ef(yp,y,xp,x)=0
% y=g(x);
% xp=h(x)+eta*ep
% h(x)=[tilh(x);Phi(x)];
%
% The functions f and Phi are known.
% The functions g and tilh are unknown.

clear,clc

% Step 1: Define the model by symbolic variables.

syms logk logkp logc logcp loga logap real
syms BETA GAMMA ALPHA RHO DELTA SIGMA real

% State variables: log capital and log technology
x=[logk,loga]; % current period
xp=[logkp,logap]; % future period

% Control variable: log consumption
y=[logc]; % current period
yp=[logcp]; % future period

% Parameters
symparams=[BETA,GAMMA,ALPHA,RHO,DELTA,SIGMA];

% Define the function Phi, which is the lower block of h(x).
% This block is the expected value of the exogenous state variables.

Phi=RHO*loga;

% Define eta as in Schmitt-Grohe and Uribe (2004):

eta=[0;SIGMA];

% Log and level of future mpk.
logmpkp=log(ALPHA)+logap+(ALPHA-1)*logkp; 
mpkp=exp(logmpkp);

% Log and level of stochastic discount factor.
logmp=log(BETA)+GAMMA*(logc-logcp);
mp=exp(logmp);

% Log and level of output.
logoutput=loga+ALPHA*logk;
output=exp(logoutput);

% Capital and consumption.
k=exp(logk);
c=exp(logc);
kp=exp(logkp);

% Budget constraint
f1=output+(1-DELTA)*k-c-kp;

% Euler condition
f2=mp*(mpkp+1-DELTA)-1;

f=[f1;f2]; % all conditions

% Choose approximation order

approx=5; % fifth order solution.

% call differentiate_dsge

model=differentiate_dsge(f,yp,y,xp,x,symparams,approx,Phi);
save('model')