% This script shows how to solve the neoclassical growh model by Smolyak
% collocation.
%
% The model is defined by:
%
% Ef(yp,y,xp,x)=0
% y=g(x);
% xp=h(x)+eta*ep
% h(x)=[tilh(x);Phi(x)];
%
% The functions f and Phi are known.
% The functions g and tilh are unknown. The algorithm will approximate them
% by Smolyak polynomials and solve the coefficients by collocation.

clear,clc

% Define the model by symbolic variables.

syms logk logkp logc logcp loga logap real
syms BETA GAMMA ALPHA RHO DELTA SIGMA real

% State variables: log capital and log technology
x=[logk,loga]; % current period
xp=[logkp,logap]; % future period

% Control variable: log consumption
y=[logc]; % current period
yp=[logcp]; % future period

% Parameters
symparams=[BETA,GAMMA,ALPHA,RHO,DELTA,SIGMA];

% Define the function Phi, which is the lower block of h(x).
% This block is the expected value of the exogenous state variables.

Phi=RHO*loga;

% Define eta as in Schmitt-Grohe and Uribe (2004):

eta=[0;SIGMA];

% Use auxiliary variables. This would speedup computation.

% Log and level of future mpk.
syms mpkp logmpkp real
logmpkp_=log(ALPHA)+logap+(ALPHA-1)*logkp; 
mpkp_=exp(logmpkp);

% Log and level of stochastic discount factor.
syms mp logmp real
logmp_=log(BETA)+GAMMA*(logc-logcp);
mp_=exp(logmp);

% Log and level of output.
syms logoutput output real
logoutput_=loga+ALPHA*logk;
output_=exp(logoutput);

% Capital and consumption.
syms k c kp real
k_=exp(logk);
c_=exp(logc);
kp_=exp(logkp);

% Define inverse capital to normalize the budget constraint (it is a good
% practice to normalize the model conditions).
syms invk real
invk_=exp(-logk);

% Budget constraint
f1=[output+(1-DELTA)*k-c-kp]*invk;

% Euler condition
f2=mp*(mpkp+1-DELTA)-1;

f=[f1;f2]; % all conditions

% all auxiliary variables and auxiliary functions:

auxvars=[mpkp;logmpkp;mp;logmp;logoutput;output;k;c;kp;invk];
auxfuns=[mpkp_;logmpkp_;mp_;logmp_;logoutput_;output_;k_;c_;kp_;invk_];

% Note that f is a function of the model variables and the auxiliary
% variables. To get f as a function of the model variables only, use the
% function subsf:

f_noaux = subsf( f,auxvars,auxfuns );

% Compare f with f_noaux

f,f_noaux

% Display the auxiliary equations:

[auxvars,auxfuns]

% Choose approximation order

order=2; % Smolyak approximation level.

% Call prepare_collocation

model=prepare_collocation(f,Phi,yp,y,xp,x,eta,symparams,order,auxfuns,auxvars);

save('model')

