clear,clc

% Add folder 'files' to the search path:
addpath('files'); 

% Load the model:
load('model')

% Choose parameter values:
BETA=.96; GAMMA=2; ALPHA=.3; RHO=.8; DELTA=.1; SIGMA=.02;

% Make a vector of parameter values:
params=eval(symparams);

% Make eta
eta=eval(eta);

% Discretize the shocks. You can use the monomial rules written by Judd, Maliar, Maliar and Valero (2014).

n_e=size(eta,2); % number of shocks.

[n_nodes,nep,P] = Monomials_2(n_e,eye(n_e)); 
nep=nep'; % transpose to n_e-by-n_nodes

% Choose Jacobian type:

model.jacobian='exact'; 
% model.jacobian='approximate'; 

% Get a perturbation solution to be used as an initial guess

% Steady state values
kss=((1/BETA-1+DELTA)/ALPHA)^(1/(ALPHA-1));
css=kss^ALPHA-DELTA*kss;
ass=1;

nxss=[log(kss);log(ass)];
nyss=log(css);

% Check that the steady state is correct:

double(subs(f_noaux,[yp(:);y(:);xp(:);x(:);symparams(:)],[nyss(:);nyss(:);nxss(:);nxss(:);params(:)]))

% Compute the cross moments:

M=get_moments(nep,P,model.order(2));

% Choose algorithm for the Sylvester equation
% algo='vectorize'; % Simple vectorization.
algo='dlyap'; % Hessenberg-Schur algorithm.
% algo='gensylv'; % Kamenik algorithm

% Compute perturbation solution (keep the 4 outputs):

[derivs,stoch_pert,nonstoch_pert,model]=get_pert(model,params,M,eta,nxss,nyss,algo);

% Explanation of outputs:
% derivs=structure with the perturbation solution as explained in Levintal
% (2015): "Fifth Order Perturbation Solution to DSGE Models".
% stoch_pert=vector of unique polynomial coefficients of the perturbation
% solution.
% nonstoch_pert=vector of unique polynomial coefficients of a perturbation
% solution of a nonstochastic model.


% Solve the model by Taylor projection using perturbation for
% the initial guess.

x0=nxss; %solve at the steady state
coeffs=stoch_pert; % this is the perturbation solution (you can also use nonstoch_pert, which is a perturbation solution of a nonstochastic model).
c0=nxss; % the center of the initial guess

% tolerance parameters for the Newton solver
tolX=1e-6;
tolF=1e-6;
maxiter=10;

disp(' ')
disp('1. Solve at the steady state by Newton method:')
disp(' ')
[ncoeffs,model]=tpsolve(coeffs,x0,model,params,eta,c0,nep,P,tolX,tolF,maxiter);
% ncoeffs=the solution in terms of unique Taylor coefficients about c0
% model=the structure model with some additional data computed during the
%       solution. Use it in subsequent calls to tpsolve.

disp(' ')
disp('2. Solve at a different point using the previous solution as the initial guess')
disp(' ')
x0=nxss+.8;
[ncoeffs]=tpsolve(ncoeffs,x0,model,params,eta,c0,nep,P,tolX,tolF,maxiter);


% The function tpsolve uses the Newton method for up to maxiter iterations. If it fails, it
% switches automatically to fsolve for another maxiter iterations. You can
% control the parameters of the second solver by optimoptions. The
% supported solvers are fsolve and lsqnonlin.

% For example, do one Newton iteration and switch to lsqnonlin:
disp(' ')
disp('3. Do one Newton iteration and then switch to lsqnonlin:')
disp(' ')
x0=nxss+1;
maxiter=1; % one Newton iteration
OPTIONS = optimoptions('lsqnonlin','TolX',tolX,'TolF',tolF,'MaxIter',10,'Display','iter-detailed'); % 10 more iterations by lsqnonlin
[ncoeffs]=tpsolve(ncoeffs,x0,model,params,eta,c0,nep,P,tolX,tolF,maxiter,OPTIONS);

% The following function performs only the Newton method.
% ncoeffs=tp_newton(coeffs,x0,model,params,eta,c0,nep,P,order,tolX,tolF,maxiter);

% Some useful functions:

% Evaluate model residuals (R), the policy functions (g,h1), the function Phi,
% and the auxiliary varialbes (auxvars):

coeffs=ncoeffs;
x0=nxss+.1; % take some state

% evaluate the model conditions at this state
[R,g,tilh,nPhi,nauxvars]=eval_model(coeffs,x0,model,params,eta,c0,nep,P);

% The auxiliary variables are computed for each current and future state.
% The dimensions are
% size(nauxvars)=[length(auxvars),1,length(P)]

% evaluate only the policy functions
[g,tilh,nPhi]=eval_policy(coeffs,x0,model,params,c0);

% Construct the function h as follows:

h=[tilh;nPhi];

% The control variables are given by:
ny=g;

% The state variables in the next period are given by:

shocks=.1; % the realized shocks in the next period
nxp=h+eta*shocks;